/*
 * cw.c
 *
 * (C) Manfred Haertel <Manfred.Haertel@rz-online.de> 1999 (under GPL license)
 *
 * CW driver for PC speaker
 *
 */

/* include our own definitions */

#include "cw.h"

/* initialization routine for cw driver */

int init_module ( void )
{
	int result ;
	
	/* announce */

	printk ( KERN_INFO "cw.o: Driver (C) Manfred Haertel 1999\n" ) ;
	printk ( KERN_INFO "cw.o: Driver V%s loading...\n" , CW_VERSION ) ;
	
	/* try to register the device and give appropriate messages */

	result = register_chrdev ( CW_MAJOR , "cw" , &cw_fops ) ;
	
	if ( result < 0 )
	{
		printk ( KERN_INFO "cw.o: Failure during load!\n" ) ;
		return result ;
	}
	else
	{
		printk ( KERN_INFO "cw.o: Driver successfully loaded.\n" ) ;
		return 0 ;
	}
}

/* cleanup routine for cw driver */

void cleanup_module ( void )
{
	int result ;
	
	/* try to unregister the device and give appropriate messages */
	
	result = unregister_chrdev ( CW_MAJOR , "cw" ) ;
	
	if ( result < 0 )
		printk ( KERN_INFO "cw.o: Failure during unload!\n" ) ;
	else
		printk ( KERN_INFO "cw.o: Driver successfully unloaded.\n" ) ;
}

/* write routine for cw_driver */

int cw_write ( struct file *file , const char *buf , size_t count ,
               loff_t *ppos )
{
	char output_string [MAXSTRING] ;
	int length ;
	struct wait_queue *wait = NULL ;
	int counter ;
	
	/* truncate string if necessary */
	
	if ( count < MAXSTRING )
	{
		length = count ;
	}
	else
	{
		length = MAXSTRING ;
	}
	
	/* copy to kernel space */
		
	copy_from_user ( output_string , buf , length ) ;
	
	/* printk ( KERN_INFO "cw.o: String is \"%s\".\n" , output_string ) ; */
	/* printk ( KERN_INFO "cw.o: String length is %d.\n" , count ) ; */
	
	/* sound all characters */

	for ( counter = 0 ; counter < length ; counter ++ )
	{
		cw_char ( output_string [counter] ) ;
	}
	
	/* tell linux how many characters were sent - if more were requested
	   than we can handle, linux simply calls this routine again! */
	
	return length ;
}

/* turn beeper on or off */

void cw_beeper ( unsigned long state )
{
	unsigned int count = 1193180 / FREQ ;
	
	/* printk ( KERN_INFO "cw.o: cw_beep schedulded.\n" ) ; */

	if ( state != 0 )
	{
		/* turn beeper on. disable interrupts while issuing the
		   commands to avoid side effects due to incomplete ios */
	
		cli () ;
		outb_p ( inb_p ( 0x61 ) | 3 , 0x61 ) ;
		outb_p ( 0xb6, 0x43 ) ;
		outb_p ( count & 0xff , 0x42 ) ;
		outb ( ( count >> 8 ) & 0xff, 0x42 ) ;
		sti () ;
	}
	else
	{
		/* turn beeper off */
	
        	outb ( inb_p ( 0x61 ) & 0xfc , 0x61 ) ;
        }

	/* printk ( KERN_INFO "cw.o: cw_beep done.\n" ) ; */
}

/* passive wait, schedule other processes meanwhile */

void cw_wait ( unsigned long wait_time )
{
	current->state = TASK_INTERRUPTIBLE ;
	schedule_timeout ( wait_time ) ;
}

/* handle single characters */

void cw_char ( char next_char )
{
	switch ( next_char )
	{
		case ' ':
			cw_very_long_wait () ;
			break ;
		case 'A':
		case 'a':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'B':
		case 'b':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'C':
		case 'c':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'D':
		case 'd':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'E':
		case 'e':
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'F':
		case 'f':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'G':
		case 'g':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'H':
		case 'h':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'I':
		case 'i':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'J':
		case 'j':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'K':
		case 'k':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'L':
		case 'l':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'M':
		case 'm':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'N':
		case 'n':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'O':
		case 'o':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'P':
		case 'p':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'Q':
		case 'q':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'R':
		case 'r':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'S':
		case 's':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case 'T':
		case 't':
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'U':
		case 'u':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'V':
		case 'v':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'W':
		case 'w':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'X':
		case 'x':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'Y':
		case 'y':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case 'Z':
		case 'z':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '':
		case '':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '':
		case '':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '':
		case '':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '1':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '2':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '3':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '4':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '5':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '6':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '7':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '8':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '9':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '0':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '.':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case ',':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '=':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait() ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '/':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			break ;
		case '?':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case ':':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '\'':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case ';':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case '(':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_long_wait () ;
			break ;
		case ')':	
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '-':
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_long_wait () ;
			break ;
		case '"':
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			cw_long_beep () ;
			cw_short_wait () ;
			cw_short_beep () ;
			cw_short_wait () ;
			break ;
	}
}

/* cw dot */

void cw_short_beep ( void )
{
	cw_beeper ( 1 ) ;
	cw_wait ( TICK ) ;
	cw_beeper ( 0 ) ;
}

/* cw line */

void cw_long_beep ( void )
{
	cw_beeper ( 1 ) ;
	cw_wait ( 3 * TICK ) ;
	cw_beeper ( 0 ) ;
}

/* wait between dots and lines */

void cw_short_wait ( void )
{
	cw_wait ( TICK ) ;
}

/* wait between characters */

void cw_long_wait ( void )
{
	cw_wait ( 3 * TICK ) ;
}

/* wait between words */

void cw_very_long_wait ( void )
{
	cw_wait ( 5 * TICK ) ;
}

/* read routine: give informations about the driver */

int cw_read ( struct file *file , char *buf , size_t count , loff_t *ppos )
{
	char output_string [MAXSTRING] ;

	/* handle each output line differently */

	switch ( (int) file->private_data )
	{
		case 1:
			sprintf ( output_string , "cw.o driver version %s\n" ,
			          CW_VERSION ) ;
			break ;
		case 2:
			sprintf ( output_string , "Author: %s\n" , AUTHOR ) ;
			break ;
		case 3:
			sprintf ( output_string , 
			          "Major number for /dev/cw: %d\n" ,
			          CW_MAJOR ) ;
			break ;
		case 4:
			sprintf ( output_string , "Tick Rate: %d\n" , TICK ) ;
			break ;
		default:
			/* if we have nothing for output, just say eof by
			   returning zero */
			   
			return 0 ;
	}
	
	/* copy to user buffer */
	
	copy_to_user ( buf , output_string , strlen ( output_string ) ) ;
		                
	/* increase output state variable */
	
	file->private_data ++ ;
	
	/* tell linux what we have done (how many characters were sent) */

	return strlen ( output_string ) ;
}

/* cw open routine */

int cw_open ( struct inode *inode , struct file *file )
{
	/* printk ( KERN_INFO "cw.o: calling cw_open...\n" ) ; */

	/* put state variable into file->private_data. don't use global
	   variables here because they are global to the whole system */

	file->private_data = (void *) 1 ;

	/* say that everything is ok */

	return 0 ;
}
